
//==================================================================
//
//  File: CLists.cpp
//
//  A simple forward-directed, singly linked list of Nodes;
//  each Node contains a single int datum.
//
//  Author:  Bary W Pollack
//  Version: Jul. 21, 2001
//  Updated: Dec. 8, 2002
//
//==================================================================

#include <iostream>
#include <cstdlib>
#include <cmath>

using namespace std;

//==================================================================
//  A single node of the list structure.
//  A Node is implemented as an inner class.
//  It contains:
//      nValue - an int,
//      pNext  - a link to the next node in the list
//==================================================================

class Node 
{
  public:
    Node(void) : nValue(0), pNext(NULL) { }
    Node(int nValue) : nValue(nValue), pNext(NULL) { }
    Node(int nValue, Node *pNext) : nValue(nValue), pNext(pNext) { }

    friend class List;
    friend ostream& operator << (ostream& outs, List& list);

  private:
    int   nValue;
    Node  *pNext;
};

//==================================================================

class List 
{
  public:
    List (void) : pHead(NULL) { }
    List (Node *node) : pHead(node) { }
    List (List& list) : pHead(ListCopy (list)) { }
    ~List (void);
    List& operator = (List rhs);
    bool IsEmpty (void) { return (pHead == NULL); }
    void Add (int nValue) { pHead = new Node (nValue, pHead); }
    void Insert (int nValue);
    int  Remove (void);
    int  Peek (void) { return (pHead != NULL) ? pHead->nValue : 0; }
    void Reverse (void);
    void Merge (List *pList);
    void InsertionSort (void);
    int  Length (void);
    friend ostream& operator << (ostream& outs, List& list)
    {
        outs << "(";
        for (Node *pNode = list.pHead; pNode != NULL; pNode = pNode->pNext)
            outs << " " << pNode->nValue;
        outs << " )";
        return outs;
    }
    friend Node *CreateList (int n);

  private:
    Node *ListCopy (List& list);
    Node *pHead;
};

//==================================================================
//  Destructor
//==================================================================

List::~List (void)
{
    Node *pNode = pHead;
    while (pHead != NULL)
    {
        pNode = pNode->pNext;
        delete pHead;
        pHead = pNode;
    }
}

//==================================================================
//  Overloaded assignment operator
//==================================================================

List& List::operator = (List rhs)
{
    if (this != &rhs)
    {
        this->~List();
        pHead = ListCopy (rhs);
    }
    return *this;
}

//==================================================================
//  Creates a new node with datum, nValue; Inserts the new node 
//  at the proper location in the list so that the list remains
//  sorted in ascending order.
//
//  nValue - The int value to be Inserted in to the list
//==================================================================

void List::Insert (int nValue)
{
    if (pHead == NULL || nValue < pHead->nValue)
        pHead = new Node (nValue, pHead);
    else
    {
        Node *p = pHead, *q = pHead->pNext;
        while (q != NULL && nValue > q->nValue)
        {
            p = q; q = q->pNext;
        }
        p->pNext = new Node (nValue, q);
    }
}

//==================================================================
//  Removes the head node of the list; returns its datum as the value.
//==================================================================

int List::Remove (void)
{
    int nValue = 0;
    if (pHead != NULL)
    {
        nValue = pHead->nValue;
        Node *node = pHead;
        pHead = pHead->pNext;
        delete node;
    }
    return nValue;
}

//==================================================================
//  Reverses the list
//==================================================================

void List::Reverse (void)
{
    List *pRevList = new List;
    while (pHead != NULL)
        pRevList->Add (Remove ());
    pHead = pRevList->pHead;
}

//==================================================================
//  Merges two sorted lists together, returns the sorted Merged result.
//==================================================================

void List::Merge (List *pList)
{
    List *pMergedList = new List;
    while (pHead != NULL || pList->pHead != NULL)
    {
        if (pHead == NULL)
        {
            pMergedList->Add (pList->pHead->nValue);
            pList->pHead = pList->pHead->pNext;
        }
        else if (pList->pHead == NULL || pHead->nValue < pList->pHead->nValue)
        {
            pMergedList->Add (pHead->nValue);
            pHead = pHead->pNext;
        }
        else
        {
            pMergedList->Add (pList->pHead->nValue);
            pList->pHead = pList->pHead->pNext;
        }
    }
    pMergedList->Reverse ();
    pHead = pMergedList->pHead;
}

//==================================================================
//  Sort the current list into ascending order 
//  using the Insertion sort algorithm.
//==================================================================
    
void List::InsertionSort (void)
{
    List *pList = new List;
    while (pHead != NULL)
        pList->Insert (Remove ());
    pHead = pList->pHead;
}

//==================================================================
//  Calculate the length of the list; return the list's length.
//==================================================================

int List::Length (void)
{
    int nLength = 0;
    for (Node *pNode = pHead; pNode != NULL; pNode = pNode->pNext)
        ++nLength;
    return nLength;
}

//==================================================================
//  Copy a list structure; return a pointer to its head
//==================================================================

Node *List::ListCopy (List& list)
{
    Node *pHead = NULL, *pTail = NULL, *pList;

    for (pList = list.pHead ; pList != NULL; pList = pList->pNext)
    {
        if (pHead == NULL)
            pHead = pTail = new Node (pList->nValue);
        else
            pTail = pTail->pNext = new Node (pList->nValue);
    }
    return pHead;
}

//==================================================================
//  Driver to demonstrate use of the List class.
//==================================================================

//==================================================================
//  Create a list containing n random int's in the range 0..99.
//
//  n - The number of items in the list
//  Returns the created list containing n random int's.
//==================================================================

double Random (void) { return (fabs (100.0 * rand()) / RAND_MAX + 1.0); }

Node *CreateList (int n)
{
    List *pList = new List;
    for (int i = 0; i < n; i++)
        pList->Add ((int) Random ());
    return pList->pHead;
}

//==================================================================
//  Create several lists of varying Lengths.
//  Demonstrate that InsertionSort() and Merge work properly.
//==================================================================

void CLists(void)
{
    List mergedList;
    srand (17231);

    for (int i = 0; i < 13; i = (i < 3) ? i + 1 : i + 3)
    {
        List list (CreateList (i));
        cout << "Unsorted: " << list.Length () << ": " << list << endl;
        list.InsertionSort ();
        cout << "Sorted:   " << list.Length () << ": " << list << endl;
        mergedList.Merge (&list);
        cout << "Merged:   " << mergedList.Length () << ": " << mergedList << endl
             << endl;
    }
}

//==================================================================

int main (void)
{
    cout << "Demonstrate CList - Add, Insert, InsertionSort, Merge..." << endl
         << endl;
    CLists ();
    cout << "*** Fin ***" << endl
         << endl;

    return 0;
}

//==================================================================


/* ..... Output from the program .....

Demonstrate CList - Add, Insert, InsertionSort, Merge...

Unsorted: 0: ( )
Sorted:   0: ( )
Merged:   0: ( )

Unsorted: 1: ( 72 )
Sorted:   1: ( 72 )
Merged:   1: ( 72 )

Unsorted: 2: ( 50 27 )
Sorted:   2: ( 27 50 )
Merged:   3: ( 27 50 72 )

Unsorted: 3: ( 23 18 73 )
Sorted:   3: ( 18 23 73 )
Merged:   6: ( 18 23 27 50 72 73 )

Unsorted: 6: ( 2 41 99 92 69 94 )
Sorted:   6: ( 2 41 69 92 94 99 )
Merged:   12: ( 2 18 23 27 41 50 69 72 73 92 94 99 )

Unsorted: 9: ( 47 86 78 68 48 70 34 87 73 )
Sorted:   9: ( 34 47 48 68 70 73 78 86 87 )
Merged:   21: ( 2 18 23 27 34 41 47 48 50 68 69 70 72 73 73 78 86 87 92 94 99 )

Unsorted: 12: ( 64 75 12 44 9 79 95 29 52 41 19 60 )
Sorted:   12: ( 9 12 19 29 41 44 52 60 64 75 79 95 )
Merged:   33: ( 2 9 12 18 19 23 27 29 34 41 41 44 47 48 50 52 60 64 68 69 70 72 73 73 75 78 79 86 87 92 94 95 99 )

*** Fin ***

..... end of output ..... */
